/********************************************************************
*
* The example demonstrates how grab is double-buffered
* to achieve real-time processing. Asynchronous grab mode is used
* so that only a single thread is needed.
*
********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <conio.h>

#include "imapi.h"

#define MAX_CAMERAS 4
#define MAX_BUFFERS 3000

/* Prototypes */
void ErrHandler(void *Param);

void main(int argc, char **argv)
{
	long Device[MAX_CAMERAS];  /* Genesis device */
	long Camera[MAX_CAMERAS];  /* Camera */
	long Thread[MAX_CAMERAS];  /* Thread to execute all functions */
	long GrabOSB[MAX_CAMERAS][2]; /* OSBs used for synchronization */
	long InBuf[MAX_CAMERAS][2];   /* Double-buffered input buffer */
	long HostOutBuf[MAX_CAMERAS][MAX_BUFFERS];     /* Double-buffered input buffer */
	long CtrlBuf[MAX_CAMERAS]; /* Control buffer */
	long SizeX, SizeY;         /* Image size */
	long NumBands;             /* Number of bands in image */
	long Niter = 0;
	long c;
	long b;
	long i;
	long Start, End;
	double fps;
	long hostIndex;
	BOOL allocDone;
	long numberOfHost[MAX_CAMERAS];
	int beginTime;	
	char fileName[1024];
	int num_frames;
	char *dcf_file;
	char full_dcf_file[1024];
	char *save_file;
	int cameras[MAX_CAMERAS];
	int num_cameras;
	
	if (argc != 5) {
		printf("lc3 cameras num_frames dcf_file save_file");
		printf("\nAuthor: Ross Cutler\n");
		printf("Date: 3/30/99\n");
		exit(0);
	}
	num_cameras=strlen(argv[1]);
	if (num_cameras > MAX_CAMERAS) {
		printf("# cameras > %d\n", MAX_CAMERAS);
		exit(0);
	}
	for (i = 0; i < num_cameras; i++) {
		c = argv[1][i] - '0';
		if (c < 0 || c >= MAX_CAMERAS) {
			printf("invalid camera: %c\n", argv[1][i]);
			exit(0);
		}
		cameras[i]=c;
	}

	num_frames = atoi(argv[2]);
	if (num_frames > MAX_BUFFERS) {
		printf("max frames = %d\n", MAX_BUFFERS);
		exit(0);
	}
	dcf_file = argv[3];
	save_file = argv[4];
	
	/* Establish an error handler */
	imAppCatchError(IM_DEFAULT, ErrHandler, NULL);
	
	/* set timeout */
	imAppControl(IM_APP_TIMEOUT,IM_INFINITE);
	
	/* display number of systems */
	printf("number of systems = %d\n", imSysInquire(0, IM_SYS_NUM_SYSTEMS, NULL));
	
	/* Allocate the device and a thread */
	for (c = 0; c < num_cameras; c++) {
		imDevAlloc(cameras[c], 0, NULL, IM_DEFAULT, &Device[c]);
		if (!Device[c])
			printf("imDevAlloc() failed on device %d\n", c);
		imThrAlloc(Device[c], 0, &Thread[c]);
	}
	
	/* Allocate a camera */
	beginTime = time(0);
	// allocate the cameras in reverse order to fix a GNL bug!
	// otherwise, allocation takes (num_cameras-1)*9 seconds.
	sprintf(full_dcf_file, "c:\\genesis\\dcf\\kodak\\es-310\\%s", dcf_file);
	for (c = num_cameras-1; c >= 0; c--)
		imCamAlloc(Thread[c], full_dcf_file, IM_DEFAULT, &Camera[c]);
	printf("time imCamAlloc() = %d seconds\n",time(0)-beginTime);
	
	/* Determine the image size and number of bands */
	imCamInquire(Thread[0], Camera[0], IM_DIG_SIZE_X, &SizeX);
	imCamInquire(Thread[0], Camera[0], IM_DIG_SIZE_Y, &SizeY);
	imCamInquire(Thread[0], Camera[0], IM_DIG_NUM_BANDS, &NumBands);
	
	/* Force grabbed image to be contigous in host memory (for imBufPut) */
	SizeX = (SizeX + 3) & ~3; 
	printf("Image size: %ix%i\n", SizeX, SizeY);
	
	/* Disable buffering of stdout */
	setvbuf(stdout, NULL, _IONBF, 0);
	
	/* Allocate grab buffers in host memory */
	for (c = 0; c < num_cameras; c++) {
		imBufAlloc(Thread[c], SizeX, SizeY, NumBands, IM_UBYTE, IM_PROC, &InBuf[c][0]);
		imBufAlloc(Thread[c], SizeX, SizeY, NumBands, IM_UBYTE, IM_PROC, &InBuf[c][1]);
	}
	
	/* Disable error catching */
	imAppCatchError(IM_DEFAULT, NULL, NULL);
	beginTime = time(0);
	
	for (c = 0; c < num_cameras; c++) {
		hostIndex = 0;
		numberOfHost[c] = 0;
		allocDone = FALSE;
		while((hostIndex < num_frames) && (allocDone == FALSE))
		{
			imBufAlloc(Thread[c], SizeX, SizeY, NumBands, IM_UBYTE, IM_HOST, &HostOutBuf[c][hostIndex]);
			if(HostOutBuf[c][hostIndex] == 0)
			{
				allocDone = TRUE;
			}
			else
			{
				hostIndex++;
				numberOfHost[c]++;
			}
		}
	}
	printf("buffer allocation took %d seconds\n", time(0) - beginTime);
	
	/* REEnable error catching */    
	imAppCatchError(IM_DEFAULT, ErrHandler, NULL);
	
	/* Allocate control buffer */
	for (c = 0; c < num_cameras; c++)
		imBufAllocControl(Thread[c], &CtrlBuf[c]);
	
	/* Allocate OSBs for synchronization */
	for (c = 0; c < num_cameras; c++) {
		imSyncAlloc(Thread[c], &GrabOSB[c][0]);
		imSyncAlloc(Thread[c], &GrabOSB[c][1]);
	}
	
	/* Select asynchronous grab mode */
	for (c = 0; c < num_cameras; c++)
		imBufPutField(Thread[c], CtrlBuf[c], IM_CTL_GRAB_MODE, IM_ASYNCHRONOUS);
	
	printf("Grabbing on %li host buffers\n",numberOfHost[0]);
	
	/* Make sure previous commands have finished before starting the timer */
	for (c = 0; c < num_cameras; c++)
		imSyncHost(Thread[c], 0, IM_COMPLETED);
	
	/* Start the timer */
	Start = clock();
	
	imAppGetError(IM_ERR_RESET, 0);
	
	/* First grab */
	for (c = 0; c < num_cameras; c++)
		imDigGrab(Thread[c], 0, Camera[c], InBuf[c][0], 1, CtrlBuf[c], GrabOSB[c][0]);
	
	b = 1;
	hostIndex = 0;
	while (!imAppGetError(IM_ERR_CODE, NULL) && hostIndex < (numberOfHost[0] - 1))
	{
		/* Queue next grab into other buffer */
		for (c = 0; c < num_cameras; c++)
			imDigGrab(Thread[c], 0, Camera[c], InBuf[c][b], 1, CtrlBuf[c], GrabOSB[c][b]);
		
		/* Switch buffers */
		b = 1 - b;
		
		for (c = 0; c < num_cameras; c++) {
			/* Process each frame as soon as the grab completes */
			imSyncHost(Thread[c], GrabOSB[c][b], IM_COMPLETED);
			
			/* Copy processed image to the host */
			imBufCopy(Thread[c], InBuf[c][b], HostOutBuf[c][hostIndex], 0, 0);
		}
		hostIndex++;
		Niter++;
	}
	
	/* frame the last frame from each camera  */
	for (c = 0; c < num_cameras; c++) {
		imSyncHost(Thread[c], GrabOSB[c][1-b], IM_COMPLETED);
		imBufCopy(Thread[c], InBuf[c][1-b], HostOutBuf[c][hostIndex], 0, 0);
	}
	
	/* Print the frame rate */
	End = clock();
	if (End - Start > CLOCKS_PER_SEC) {
		fps = Niter * CLOCKS_PER_SEC / (double) (End - Start);
		printf("Frames/sec = %.1f\n", fps);
		Start = clock();
		Niter = 0;
	}
	
	/* save images */
	beginTime = time(0);
	for (c = 0; c < num_cameras; c++) {
		for(b=0;b<numberOfHost[c];b++) {
			sprintf(fileName, "%s.%d.%04d.tif", save_file, cameras[c], b);
			imBufSave(Thread[c], fileName, IM_TIFF, HostOutBuf[c][b]);
		}
	}
	printf("save images: %d seconds\n",time(0)-beginTime);
		
	/* Clean up */
	for (c = 0; c < num_cameras; c++)
		for(b=0;b<numberOfHost[c];b++)
			imBufFree(Thread[c], HostOutBuf[c][b]);
		
	for (c = 0; c < num_cameras; c++) {
		imBufFree(Thread[c], InBuf[c][0]);
		imBufFree(Thread[c], InBuf[c][1]);
		imBufFree(Thread[c], CtrlBuf[c]);
		imSyncFree(Thread[c], GrabOSB[c][0]);
		imSyncFree(Thread[c], GrabOSB[c][1]);
		imCamFree(Thread[c], Camera[c]);
		imThrFree(Thread[c]);
		imDevFree(Device[c]);
	}
}


void ErrHandler(void *Param)
{
	char Error[IM_ERR_SIZE];
	/*
	* Get the error message and print it. Don't reset the error
	* because we only want the first to be printed.
	*/
	imAppGetError(IM_ERR_MSG_FUNC, Error);
	printf("%s\n", Error);
}

